/*
 * Turn on MASC tty discipline on a given tty
 * $Id: ttymasc.c,v 1.10 1999/07/20 00:26:42 kaz Exp $
 * $Name: masc_2_5 $
 */

#include <termios.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include "mobitex_masc_ioctl.h"

#ifndef N_MASC
#define N_MASC 8
#endif

static char *progname;

static void progerr(char *msg)
{
	fprintf(stderr, "%s: ", progname);
	perror(msg);
}

static void tweak_tty_params(struct termios *ti, speed_t speed)
{
	ti->c_iflag = INPCK | IGNBRK;
	ti->c_oflag = 0;
	ti->c_cflag = CS7 | CREAD | PARENB | HUPCL | CLOCAL | CRTSCTS;
	ti->c_lflag = 0;
	cfsetospeed(ti, speed);
	cfsetispeed(ti, speed);
}

static int get_tty_params(int fd, struct termios *old)
{

	if (tcgetattr(fd, old) < 0) {
		progerr("tcgetattr()");
		return 0;
	}
	return 1;
}

static int set_tty_params(int fd, const struct termios *new)
{
	if (tcsetattr(fd, TCSANOW, new) < 0) {
		progerr("tcsetattr()");
		return 0;
	}
	return 1;
}

static int set_discipline(int fd, int mascdisc)
{
	if (ioctl(fd, TIOCSETD, &mascdisc) < 0) {
		progerr("ioctl(TIOCSETD)");
		return 0;
	}
	return 1;
}

static int turnoff_modem(int fd)
{
	if (ioctl(fd, MASCIOCTURNOFF) < 0) {
		progerr("ioctl(MASCIOCTURNOFF)");
		return 0;
	}
	return 0;
}

static int transmit_on(int fd)
{
	if (ioctl(fd, MASCIOCTXON) < 0) {
		progerr("ioctl(MASCIOCTXON)");
		return 0;
	}
	return 1;
}

static int transmit_off(int fd)
{
	if (ioctl(fd, MASCIOCTXOFF) < 0) {
		progerr("ioctl(MASCIOCTXOFF)");
		return 0;
	}
	return 1;
}

static int set_network_id(int fd, unsigned long netid)
{
	if (ioctl(fd, MASCIOCSNETID, &netid) < 0) {
		progerr("ioctl(MASCIOCSNETID)");
		return 0;
	}
	return 1;
}

static int drain_wait(int fd)
{
	if (ioctl(fd, MASCIOCDRAIN) < 0) {
		progerr("ioctl(MASCIOCDRAIN)");
		return 0;
	}
	return 1;
}

static void interrupt(int sig)
{
}

static char *helptext;
static char *extra;

int main(int argc, char **argv)
{
	int fd = -1;
	int exstatus = EXIT_FAILURE;
	speed_t speed = B9600;
	unsigned long netid = 0;
	char *ttname;
	struct termios saved_params, new_params;
	sigset_t blocked;
	enum { setup_ldisc, tran_off, tran_on, set_netid } oper = setup_ldisc;
	int dont_turn_off = 0;

	if (argc == 2 && (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "--help") == 0)) {
	    	fputs(helptext, stdout);
		exstatus = EXIT_SUCCESS;
		goto exit;
	}

	if (argc == 2 && (strcmp(argv[1], "-H") == 0 || strcmp(argv[1], "--extra-help") == 0)) {
	    	fputs(extra, stdout);
		exstatus = EXIT_SUCCESS;
		goto exit;
	}

	if (argc < 2) {
		fprintf(stderr, "usage: %s <tty_name> [ <options> ]\n", argv[0]);
		goto exit;
	}

	progname = argv[0];
	ttname = argv[1];

	argc -= 2;
	argv += 2;

	while (*argv && argv[0][0] == '-') {
		int i;
		for (i = 1; i < strlen(argv[0]); i++) {
			switch(argv[0][i]) {
			case 'n':
				dont_turn_off = 1;
				break;
			case '1':
				oper = tran_on;
				break;
			case '0':
				oper = tran_off;
				break;
			case 'i':
				oper = set_netid;
				if (argv[1] == 0) {
					fprintf(stderr, "%s: option -i requires argument\n", progname);
					goto exit;
				} else {
					char *err;
					netid = strtoul(argv[1], &err, 0);

					if (*err != 0) {
						fprintf(stderr, "%s: badly formed network id\n", progname);
						goto exit;
					}
				}
				argv++;
				argc--;
				goto next_arg;
			case 's':
				if (argv[1] == 0) {
					fprintf(stderr, "%s: option -s requires argument", progname);
					goto exit;
				} else {
					char *err;
					unsigned long sp = strtoul(argv[1], &err, 0);
					if (*err != 0) {
						fprintf(stderr, "%s: badly formed speed\n", progname);
						goto exit;
					}
					switch (sp) {
					case 1200:
						speed = B1200;
						break;
					case 2400:
						speed = B2400;
						break;
					case 4800:
						speed = B4800;
						break;
					case 9600:
						speed = B9600;
						break;
					case 19200:
						speed = B19200;
						break;
					case 38400:
						speed = B38400;
						break;
					default:
						fprintf(stderr, "%s: invalid speed\n", progname);
						goto exit;
					}
				}
				argv++;
				argc--;
				goto next_arg;
			default:
				fprintf(stderr, "%s: unknown option -%c\n",
					 progname, (int) argv[0][i]);
				goto exit;
			}

		}
	next_arg:
		argv++;
		argc--;
	}

	fd = open(ttname, O_RDWR | O_NOCTTY | O_NONBLOCK);

	if (fd == -1) {
		progerr(ttname);
		goto exit;
	}

	signal(SIGTERM, interrupt);
	signal(SIGINT, interrupt);

	if (!get_tty_params(fd, &saved_params))
		goto close_exit;

	/*
	 * Only the line discipline setting operation sets up the tty
	 * parameters. The other operatons assume a tty with MASC discipline
	 * and correct tty parameters already set up.
	 */

	if (oper == setup_ldisc) {
		new_params = saved_params;
		tweak_tty_params(&new_params, speed);
		if (!set_tty_params(fd, &new_params))
			goto close_exit;
		if (!set_discipline(fd, N_MASC))
			goto close_exit;
	} else {
		if (saved_params.c_line != N_MASC) {
			fprintf(stderr, "%s: requires tty with MASC discipline enabled!\n", progname);
			goto close_exit;
		}
	}
	
	sigemptyset(&blocked);

	switch (oper) {
	case setup_ldisc:
		/*
		 * We must keep the tty open until interrupted by signal.
		 * The drain_wait is needes to ensure that the ack to
		 * the modem's F O frame is sent out to the tty! Without
		 * the ack, the modem won't complete the shutdown sequence.
		 */

		sigsuspend(&blocked);
		if (!dont_turn_off) {
			if (turnoff_modem(fd) && drain_wait(fd))
				exstatus = EXIT_SUCCESS;
		} else {
			exstatus = EXIT_SUCCESS;
		}
		break;
	case tran_on:
		if (transmit_on(fd))
			exstatus = EXIT_SUCCESS;
		break;
	case tran_off:
		if (transmit_off(fd))
			exstatus = EXIT_SUCCESS;
		break;
	case set_netid:
		if (set_network_id(fd, netid))
			exstatus = EXIT_SUCCESS;
		break;
	}

	/*
	 * Only restore parameters if we are the default ldisc setting
	 * operation.
	 */

	if (oper == setup_ldisc) {
		if (!set_discipline(fd, N_TTY))
			goto close_exit;
		set_tty_params(fd, &saved_params);
	}
close_exit:
	close(fd);
exit:
	return exstatus;
}

static char *helptext = {
	"\n"
	"ttymasc command syntax: \n"
	"\n"
	"\tttymasc <ttyname> [ -n ] [ -s ]\n"
	"\tttymasc <ttyname> [ -0 | -1 | -i N ]\n"
	"\n"
	"options:\n"
	"\n"
	"\t-H\t\textra help\n"
	"\t-n\t\tdo not turn off modem upon termination\n"
	"\t-s N\t\tset input and output baud rate to N.\n"
	"\t\t\tN may be 1200, 2400, 4800, 9600, 19200 or 38400\n"
	"\t-0\t\tturn off modem's transmitter and exit\n"
	"\t-1\t\tturn on modem's transmitter and exit\n"
	"\t-i N\t\tset 32 bit network id to N and exit.\n"
	"\t\t\tN may be written in hex, like 0x1234ABCD.\n"
	"\n"
};

static char *extra = {
	"\n"
	"ttymasc extra hints: \n"
	"\n"
	"The ttymasc program has two modes of operation: a watcher mode, and a parameter\n"
	"setting mode.\n"
	"\n"
	"In the watcher mode, what ttymasc does is open the given tty line, turns on the\n"
	"MASC (Mobitex Asynchronous Serial Communication) protocol on the line, and then\n"
	"waits to be killed, either by an interrupt from the terminal or an explicit\n"
	"kill.  When it is killed, it optionally turns the modem off, then waits for all\n"
	"outgoing high-priority packets to drain and removes the protocol from the\n"
	"serial line. Finally, it closes the tty.\n"
	"\n"
	"In the parameter setting mode, ttymasc assumes that there is already another\n"
	"ttymasc running in watcher mode on the tty, and therefore that the MASC\n"
	"protocol is running. In this mode, ttymasc executes some request to change\n"
	"modem parameters. Only one request is handled per invocation of ttymasc.\n"
	"\n"
	"If ttymasc is not given any parameters other than the tty name, it is\n"
	"in watcher mode. Also, in the watcher mode, the -s parameter may be used to\n"
	"establish the serial line input and output baud rate. The default is 9600 baud.\n"
	"The -n parameter should be specified if it is undesirable for ttymasc to turn\n"
	"off the modem when it is killed.\n"
	"\n"
	"If options other than -n or -s are given, then ttymasc is in parameter setting\n"
	"mode. In this mode, the -n or -s options, if specified, are ignored.  Also,\n"
	"only the last option is processed and its action executed. So that if for\n"
	"example -0 -1 are specified, the effect is as though only -1 were specified.\n"
	"\n"
};
